"""Callback Objects Using Weak References

    A common problem in implementing the "Observer" pattern in Python is that
    containers often want to be an observer of their contents.  This makes
    it tricky to pass callback information to the observed object without
    creating a circular reference.  In limited applications of the observer
    pattern with relatively static observer-subject relationships, one can
    code around this, but it would be much nicer if you could just pass a
    callback object to anything that needed "subscribing" to, and not have to
    worry about the created references.

    The 'TW.Callbacks' module defines a simple interface, 'ICallback', and
    provides a 'Callback' class which implements it.  It also provides a
    'CallbackList' class which manages a set of subscribed 'ICallback'
    implementations.

    The supplied implementations address a variety of low-level issues such
    as handling callbacks to objects which do not support weak references
    (in which case a normal reference is used).  There is also an experimental
    'Dispatcher' class which is intended to support event keys, with each
    callback capable of being subscribed to more than one event at once, but
    receiving only one call even if multiple simultaneous events are fired.

    Quick usage overview::

        from TW.Callbacks import Callback, CallbackList
        
        class someSubject:

            def __init__(self):
                self.observers = CallbackList()
                self.subscribe = self.observers.subscribe
                self.unsubscribe = self.observers.remove

        class anObserver:
            def __init__(self, aSubject):
                aSubject.subscribe( Callback(self,'myCallbackMethod') )

    See the classes and source code for more details.
"""

from weakref import ref
from UserList import UserList
import Interface

__all__ = 'ICallback', 'Callback','CallbackList', 'Dispatcher'


try:
    from Acquisition import aq_base
    
except ImportError:
    # If there's no acquisition, there can't be any wrappers to remove, right?
    # XXX In future, need to support Zope 3 wrapper protocols...
    def aq_base(obj): return obj



class ICallback(Interface.Base):
    """Encapsulate a callback with support for weak references
    
        CallbackList expects its subscribers to adhere to this interface.
    """
    
    def isValid(self):
        """Return false if callback is no longer valid"""

    def hasTarget(self, object):
        """Return true if callback is 'object' or has 'object' as its target"""

    def __call__(self, *args, **kw):
        """Make the callback, passing in *args and **kw as well"""

    def __cmp__(self,other):
        """Callbacks must support equality testing, even if they are only
        equal to themselves"""
    
    def __hash__(self):
        """Callbacks must be hashable"""

    

class Callback:

    """Encapsulate a callback with support for weak references"""
    
    __implements__ = ICallback
    
    def __init__(self, object, methodName='callBack', *msgArgs, **msgKw):

        try:
            self._getObject = ref(object)            
        except TypeError:
            self.object = object

        self.info = methodName, msgArgs, msgKw


    def _getObject(self):
        return self.__dict__['object']

    
    def isValid(self):
        """Return false if subscription is no longer valid"""
        return self._getObject() is not None


    def hasTarget(self, object):
        """Return true if subscription is 'object' or has 'object' as its target"""
        
        if object is self:
            return 1
        
        return aq_base(self._getObject()) is aq_base(object)


    def __cmp__(self,other):
        return cmp(id(self), id(other))
    
    def __hash__(self):
        return hash(id(self))


    def __call__(self, *args, **kw):
        """Make the callback, passing in *args and **kw as well"""
        
        obj = self._getObject()
        if obj is None: return
            
        methodName, msgArgs, msgKw = self.info
        
        kw.update(msgKw)
        args=args+msgArgs
            
        getattr(obj,methodName)(*args,**kw)





























class CallbackList(UserList):
    
    """A list of callbacks that can be called all at once"""

    def __call__(self, *args, **kw):
        for cb in self.validCallbacks():
            cb(*args, **kw)


    def validCallbacks(self):
        """Purge invalid callbacks and return only the valid ones"""
        data = self.data[:] = [cb for cb in self.data if cb.isValid()]
        return data


    def remove(self, item):
        """Remove callbacks by target; also purges invalid callbacks"""
        self.data[:] = [cb for cb in self.data if cb.isValid() and not cb.hasTarget(item)]


    def subscribe(self, callback):
        """Add a callback (synonym for append())"""
        self.append(callback)


















class Dispatcher:

    """Dispatches events by event keys"""

    # XXX No unit tests yet...  need to add some before using this
    
    def __init__(self):
        self.callbacks = {}

        
    def subscribe(self, callback, events=('*',)):
        cbs = self.callbacks
        have = cbs.has_key
        for e in events:
            if have(e):
                cbs[e].append(callback)
            else:
                cbs[e]=CallbackList([callback])

                
    def unsubscribe(self, object, events=('*',)):
        cbs = self.callbacks
        have = cbs.has_key        
        for e in events:
            if have(e):
                cbs[e].remove(object)















    def __call__(self, events=('*',), *args, **kw):
        from kjbuckets import kjSet
        targets = kjSet(20)
        member = targets.member
        add = targets.add
        args = (events,)+args

        cbs = self.callbacks
        have = cbs.has_key        
        e = kjSet(events); e.add('*')

        for e in e.values():
            if have(e):
                for cb in cbs[e].validCallbacks():
                    if not member(cb):
                        targets.add(cb)
                        cb(*args, **kw)
