"""Managed Connections to Databases, and query result helper classes

    This module supplies 'ManagedConnection' classes for a variety of
    Python database interfaces, including both SQL and LDAP.  Managed
    Connections automatically handle various matters relating to staying
    open (automatic close/re-open on errors) and participating in
    transactions.

    Currently, managed connections use the ZPublisher 'get_transaction()'
    convention for accessing the transaction, and should probably be
    updated to work with the 'TransactionManager' Service.  This needs
    some careful work, though, not to impose read-write transaction
    requirements for database connections which are in transactions by
    default.  :(

    Oh, and by the way, I haven't done anything useful with this module
    in a few months...  it may not work at all right now.  :(
"""

from TW.API import *





















class ManagedConnection(SEF.Service):

    _connection = None
    _closeASAP  = 0
    _registered = 0


    def connection(self):

        if self._closeASAP and not self._registered:
            self._close()

        conn = self._connection

        if conn is None:
            conn = self._connection = self._open()

        return conn


    def closeASAP(self):

        self._closeASAP = 1

        if not self._registered:
            self._close()


    def __del__(self):
        try:
            self.close()
        except:
            pass








    def begin(self):
        self._registered = 1    # extend in subclasses
    
    def commit(self):
        self._registered = 0    # extend in subclasses

    def abort(self):
        self._registered = 0    # extend in subclasses


    def ensure_transacted(self):

        if not self._registered:

            self.begin()
            self._registered = 1
        
            try:    t = get_transaction()
            except: pass
            else:   t.register(self.Keeper(self))
        

    class Keeper:
        def __init__(self,conn):
            self.conn = conn
        
        def tpc_begin(self,*args): pass
        def tpc_abort(self,*args): pass
        def commit(self,*args):    pass

        def tpc_finish(self,*args):
            try: self.conn.commit()
            finally: self.conn._registered = 0
        
        def abort(self,*args):
            try: self.conn.abort()
            finally: self.conn._registered = 0


    

class QueryResult:

    """May be broken - needs testing since removal of ExtensionClass.Base"""
    
    def __init__(self, list, **kw):
        self.data = list or []
        self.__dict__.update(kw)

    def __getitem__(self,item):
        return self.data[item]

    def __getslice__(self,i,j):
        self.realize(); return self.data[i:j]

    def __len__(self):
        return len(self.data)
    
    def reverse(self):
        self.data.reverse()

    def __repr__(self):
        self.realize(); return repr(self.data)
    
    def __add__(self, list):
        self.realize(); return self.data+list[:]
    
    def sort(self, *args):
        self.realize(); apply(self.data.sort, args)

    def realize(self):      
        self.data = list(self); self.realize = self.__class__.o__realize__
        self.__getitem__, self.__getslice__ = self.__class__.o__getitem__, self.__class__.o__getslice__
        
    def o__getitem__(self,item): return self.data[item]
    def o__getslice__(self,i,j): return self.data[i:j]
    def o__realize__(self): pass





class LDAPResult(QueryResult):

    def __getitem__(self,item):
        dn, item = self.data[item]
        item['dn']=dn
        return item     


class SQLResult(QueryResult):

    def __init__(self, list, desc, **kw):
        self.data = list or []
        self.Description = desc
        self.__dict__.update(kw)

    def __getitem__(self,item):
        d = {}
        item = self.data[item]
    
        for k,v in zip(self.Description,item):
            d[k[0]] = v
        
        return d


















class LDAPConnection(ManagedConnection):

    maxRetries  = 3
    resultMaker = LDAPResult

    URL = "ldap://localhost/cn=monitor"

    def _open(self):
        from AppUtils.Connections import Connect
        return Connect(self.URL)


    def _close(self):

        if self._connection is not None:
            try:
                self._connection.unbind_s()
            except:
                pass
            
            del self._connection

        if self._closeASAP:  del self._closeASAP
        if self._registered: del self._registered


    def retryMethod(self, methodName, *args, **kw):

        import ldap

        for r in range(self.maxRetries):
    
            method = getattr(self.connection(),methodName)
        
            try:    return method(*args,**kw)
            except: self._close()
            
        else:
            raise

    
    def __call__(self, baseDN=None, scope=None, filter=None, attrs=None,
                 dn_only=0, timeout=-1, resultMaker=None):

        conn = self.connection()

        if baseDN is None: baseDN = conn.basedn
        if scope is None:  scope  = conn.scope
        if filter is None: filter = conn.filter or 'objectclass=*'
        if attrs is None:  attrs  = conn.attrs or None

        results = self.retryMethod('search_st',
            baseDN, scope, filter.strip(), attrs, dn_only, timeout
        )
    
        if resultMaker is None:
            resultMaker = self.resultMaker
    
        return resultMaker(results)
    






















class SQLConnection(ManagedConnection):

    resultMaker = SQLResult

    def _close(self):
        
        if self._connection is not None:
            try:
                self._connection.close()
            except:
                pass
            
            del self._connection

        if self._closeASAP:  del self._closeASAP
        if self._registered: del self._registered


    def __call__(self,sql,rows=0,getResults=1,transacted=0,resultMaker=None):

        if not self._registered and (transacted or self._alwaysTransacted):
            self.ensure_transacted()

        try:
            desc, results = self._query(sql,getResults,rows)
        except:
            self._closeASAP = 1
            raise

        if resultMaker is None:
            resultMaker = self.resultMaker
    
        return resultMaker(results,desc,sql=sql)








    def format(self,value,fieldType='VARCHAR',precision=None,quote=1):

        """Format an atomic value for insertion into an SQL statement

            'fieldType' specifies the target datatype, 'precision' is the
            number of bytes, characters, or digits of the specified type.
            'quote' is a flag that can be set to a false value to avoid
            quoting of textual values (useful when assembling a text
            value from multiple parts).

            Subclasses can extend this method to support additional field
            types."""

        value=str(value)
    
        if fieldType in ('TEXT','CHAR','VARCHAR'):
    
            if precision is not None:
                value=(value+' '*precision)[:precision]
            
            value = value.replace("'","''").replace('\0','')
        
            if quote:
                value="'%s'" % value

        return value















class DBI(SQLConnection):

    _alwaysTransacted = 1


    def commit(self):
        self._connection.commit()
        self._registered = 0
        if self._closeASAP: self._close()

    
    def abort(self):
        self._connection.rollback()
        self._registered = 0
        if self._closeASAP: self._close()


    def _query(self,sql,getResults,rows):

        cursor = self.connection().cursor()
        cursor.execute(sql)
    
        if getResults:
    
            if rows:
                result = cursor.fetchmany(rows)
            else:
                result = cursor.fetchall()
                desc = cursor.description
        else:
            desc,result = [],[]
            
        cursor.close()
        return desc,result







class gadflyDB(DBI):

    def _open(self):

        import gadfly, os

        dbname, dbdir = self.Database,self.Directory
    
        db = gadfly.gadfly(dbname, dbdir)
    
        if not exists(os.path.join(dbdir, dbname+'.gfd')):
            db.startup(dbname, dbdir)
        
        return db


class gadflyClient(DBI):

    def _open(self):
        from gfclient import gfclient
        return gfclient(self.User,self.Port,self.Password,self.Server)




















class ctSybase(DBI):

    _alwaysTransacted = 0

    def _open(self):
        parms, database = self._connParms()
        import ctsybase, socket, os
        parms['hostname']=parms.get('hostname', socket.gethostname())
        parms['application']=parms.get('application', str(os.getpid()))
        db = ctsybase.ctsybase(parms)
        db.execute('use %s' % database)
        return db


    def begin(self):
        try:
            self.connection().execute('begin tran')
        except:
            self.closeASAP()
            raise
        super(ctSybase,self).begin()


    def _query(self,sql,getResults,rows):

        conn = self.connection()
        from ctsybase import error
    
        try:
            conn.execute('SET ROWCOUNT %d' % rows)
            result = conn.execute(sql)
        
            if result:
                return conn.description[0], result[0]
            else:
                return [],[]
        
        except error,v:
            raise error,(sql,v),sys.exc_traceback


    def _connParms(self):
        return {
            'user':self.User, 'password':self.Password, 'server':self.Server
            }, self.Database


class SolutionsSybase(ctSybase):

    def _connParms(self):

        from Solutions.Sybase import URL
        url = URL(self.URL)
    
        return {
            'user':url.user, 'password':url.password, 'server':url.server
        }, url.database


setupModule()
