"""Components for database record handling, caching, etc.

    Example usage::

        class myDatabase(DataModel.Database):
        
            class Family1(DataModel.TypeManager):
            
                class SomeTypeA1(DataModel.RecordType):
                    allKeys = (
                        ('primary_key',),
                        ('secondary_key_field1','secondary_key_field2'),
                        ...
                    )


                class SomeTypeB1(SomeTypeA1):
                    ...
                    
                class SomeTypeC1(DataModel.RecordType):
                    ...
                    
            class Family2(DataModel.TypeManager):
                ...

    You may nest as many families in a database as you want, and as many
    record types in a family as you want.  All record types in a family
    share the same class for record objects, and the same record cache,
    indexed by their combined key fields.  Record types may derive from
    one another, within the same type family.
"""










from TW.API import *
from kjbuckets import *
from TW.Database.Interfaces import *
from TW.Database.Interfaces import __all__ as allInterfaces
from TW.Callbacks import CallbackList
from weakref import WeakValueDictionary

__all__ = [
    'Database', 'Record', 'RecordType', 'TypeManager', 'SimpleTypeManager',
]

__all__ += allInterfaces   # Export interfaces too...





























class Database(SEF.Service):

    """A database; handles update queue and transactions"""

    isFlushing = 0
    
    def __init__(self):
        self.updateQueue = []
        
    def flushUpdates(self):
        assert not self.isFlushing,"Recursive attempt to flush updates"
        self.isFlushing = 1
        try:
            for record in self.updateQueue:
                record.flushWrites()
            self.updateQueue = []                                
        finally:
            del self.isFlushing

    def queueForUpdate(self, record):
        self.updateQueue.append(record)


    def _register(self, record):
        """Register record with the transaction manager"""
        record._p_jar = self
        self.TransactionManager.register(record)

    TransactionManager = SEF.bindTo('TransactionManager')












    ### Transaction Support (Zope-style transactions)

    def tpc_begin(self, transaction):
        self.flushUpdates()
        self.committedRecords = []

    def commit(self, record, transaction):
        if record.dirtyFields: record.flushWrites()
        self.committedRecords.append(record)
        del record.ensureRegistered
        
    def abort(self, record, transaction):
        record.invalidate()
        del record.ensureRegistered

    def tpc_vote(self, transaction):
        """Last chance to raise an error - no-op by default"""

    def tpc_abort(self, transaction):
        for record in self.commitedRecords:
            record.invalidate()
        self.commitedRecords = []

    def tpc_finish(self, transaction):
        """Can be overridden to do post-commit operations"""
        self.commitedRecords = []















class Record:

    """A record of a RecordType; handles field read/write"""
    
    isValid = 1


    def __init__(self, db, typeManager, **kw):
        
        self.typeManager  = typeManager
        self.keyFields    = typeManager.keyFieldNames
        self.rootTypeName = typeManager.rootTypeName
        self.db           = db
        self.data         = kw.copy()
        self.cachedByKeys = kjSet()            
        self.dirtyFields  = kjSet()

        self.typeMap           = {}
        self.mostSpecificTypes = []
        
        self.invalidationCallbacks = CallbackList()


    def exists(self):
        """Does record exist (i.e., does it include the base type)?"""
        return self.hasType(self.rootTypeName)


    def create(self):
        """Create object (by adding base type)"""
        self.addType(self.rootTypeName)


    def delete(self):
        """Delete object (by removing base type)"""
        self.removeType(self.rootTypeName)





    def updateTypeMap(self, typeObject, valid=1):

        tm  = self.typeMap
        tm[typeObject._componentName] = valid

        if valid:
        
            for t in typeObject.allTypeNames.values():
                tm[t]=valid

            self._updateMSTlist(typeObject)
            
                
        else:
            for t in typeObject.allSubtypeNames.values():
                tm[t]=valid
                
            self.mostSpecificTypes = []
            getType = self.typeManager._getType
            update  = self._updateMSTlist

            # Rebuild MST's from typemap
            for t,v in tm.items():
                if v: update(getType(t))


    def _updateMSTlist(self, typeObject):
    
        mst = self.mostSpecificTypes
        
        for t in mst[:]:
            if typeObject.isSubTypeOf(t):
                mst.remove(t)   # remove superclasses of new type
            elif t.isSubTypeOf(typeObject):
                return          # cancel if new type is super of existing
        else:
            mst.append(typeObject)  # only add if a subtype isn't present




    def markWritten(self, fields):
        df = self.dirtyFields
        for field in fields:
            del df[field]


    def flushWrites(self):
    
        for t in self.mostSpecificTypes:
            t._saveRecord(self)
            
        self.dirtyFields.clear()            
        del self.ensureQueued       # error if flush called when not queued!


    def loadDefaults(self, mapping={}, **kw):

        data = {}
        data.update(mapping)
        data.update(kw)

        have = self.data.has_key

        for k in data.keys():
            if have(k): del data[k]

        self.loadData(data)


    def loadData(self, mapping={}, **kw):
    
        data = {}
        data.update(mapping)
        data.update(kw)
        
        self.data.update(data)

        if self.keyFields & kjSet(data.keys()):
            self.typeManager._cacheRecord(self)


    def hasType(self, typeName):        
        """Does record include type 'typeName'?"""
        
        have = self.typeMap.get(typeName, NOT_GIVEN)

        if have is NOT_GIVEN:
            self.typeManager._loadDataFor(self, typeName)
            try:
                return self.typeMap[typeName]
            except KeyError:
                raise TypeError, ("%s._loadDataFor() did not updateTypeMap()" % typeName)

        return have


    def addType(self, typeName):
        """Add type 'typeName' to record"""
        assert not self.hasType(typeName)
        self.setType(typeName)


    def removeType(self, typeName):
        """Remove type 'typeName' from record"""
        assert self.hasType(typeName)
        self.unsetType(typeName)
    

    def setType(self, typeName):
        """Ensure that record includes type 'typeName'"""
        if not self.hasType(typeName):
            self.typeManager._setTypeFor(self, typeName)            


    def unsetType(self, typeName):
        """Ensure that record does not include type 'typeName'"""
        if not self.hasType(typeName):
            self.typeManager._unsetTypeFor(self, typeName)            




    def invalidate(self):
        try:
            self.invalidationCallbacks(self)
        finally:
            # Ensure that any attempt to use will cause errors
            del self.data, self.typeMap 
            self.isValid  = 0
            # And remove ourself from the cache
            self.typeManager._cacheRecord(self)

    def addInvalidationCallback(self, callback):
        self.invalidationCallbacks.append(callback)


    def removeInvalidationCallback(self, target):
        self.invalidationCallbacks.remove(callback)


    def ensureRegistered(self, d, n):
        # XXX should check version info here and abort if outdated
        self.db._register(self)
        return 1
        
    ensureRegistered = Once(ensureRegistered)


    def ensureQueued(self, d, n):
        self.db.queueForUpdate(self)
        return 1

    ensureQueued = Once(ensureQueued)










    def __getitem__(self,key):
        return self.data[key]


    def __setitem__(self, key, value):

        oldValue = self.data[key]

        if oldValue != value:
        
            newValue = value
            
            for t in self.mostSpecificTypes:
                newValue = t._changedField(self, key, oldValue, newValue)

            if newValue is not oldValue:
            
                self.ensureRegistered   # ensure registration
                self.dirtyFields.add(key)
                self.data[key] = newValue
                
                if self.keyFields.member(key):
                    self.typeManager._cacheRecord(self)


    def get(self,key,default=None):
        return self.data.get(key,default)
    
    def keys(self):     return self.data.keys()
    def values(self):   return self.data.values()
    def items(self):    return self.data.items()
    def __len__(self):  return len(self.data)









class RecordType(SEF.Service):

    """A kind of database record; knows how to load/save/query"""

    __implements__  = ITypeImpl

    allKeys = ()

    def __class_init__(thisClass, newClass, next):
        
        if __proceed__ is not None:
            __proceed__(thisClass, newClass, next)
        else:
            next().__class_init__(newClass,next)

        if thisClass is newClass:
            return

        name = newClass.__name__
        
        newClass.directSubtypeNames = []
        newClass.allSubtypeNames = kjSet([name])
        newClass.allTypeNames = kjSet([name])
        
        for base in newClass.__bases__:
            getattr(base,'directSubtypeNames',[]).append(name)

        for base in newClass.__mro__:
            # Stop when we reach ourself (thus skipping 'object' as well)
            if base is thisClass: break
            base.allSubtypeNames.add(name)
            newClass.allTypeNames.add(base.__name__)


    def isSubtypeOf(self, otherType):
        myName, hisName = self._componentName, otherType._componentName
        return hisName!=myName and self.allTypeNames.member(hisName)
        



    # Stuff that should/can be overridden to implement record type

    def _loadDataFor(self, record, typeName=None):
    
        """Define in subclasses to read data into record"""
        
        # Note: to load the data, should call:
        # record.loadData(fieldMapping); record.updateTypeMap(self,1)
        
        raise NotImplementedError


    def _setTypeFor(self, record, typeName=None):
        raise NotImplementedError


    def _unsetTypeFor(self, record, typeName=None):
        raise NotImplementedError


    def _saveRecord(self, record):
        """Define in subclasses to write out changes made to record"""
        raise NotImplementedError


    def _changedField(self, record, fieldName, oldValue, newValue):
        record.ensureQueued     # ensure record is queued for update    
        return newValue


    def getItem(self, **keyMapping):
        """Look up record using keyword arguments; returns None if not found"""
        record = self.getService('_getFromCache')(keymapping)
        if record.hasType(self.typeName): return record

    def typeName(self, d, n):
        return self.__class__.__name__
        
    typeName = Once(typeName)
    

class TypeManager(SEF.Service):

    """A family of similar recordtypes; handles caching"""

    __implements__ = ITypeManager
    
    def __class_init__(thisClass, newClass, next):

        if __proceed__ is not None:
            __proceed__(thisClass, newClass, next)
        else:
            next().__class_init__(newClass,next)

        if thisClass is newClass: return
        
        reg = []; add = reg.append
        for attName, attVal in newClass.__dict__.items():
            for newKey in getattr(attVal,'allKeys',()):
                if newKey not in reg: add(newKey)
                
        newClass.allKeys = tuple(reg)
        newClass.keyFieldNames = kjSet(
            [field for key in reg for field in key]
        )


    def _loadDataFor(self, record, typeName):
        self._getType(typeName)._loadDataFor(record)

    def _setTypeFor(self, record, typeName):
        self._getType(typeName)._setTypeFor(record)

    def _unsetTypeFor(self, record, typeName):
        self._getType(typeName)._unsetTypeFor(record)







    def _cacheRecord(self, record):
    
        oldKeys = record.cachedByKeys
        if record.isValid:
            newKeys = kjSet(self._getKeysFor(record))
        else:
            newKeys = kjSet()
            
        cache   = self.recordCache
        get     = cache.get

        for key in (oldKeys - newKeys).values():
            if get(key) is record: del cache[key]
            
        for key in (newKeys - oldKeys).values():
            cache[key] = record

        record.cachedByKeys = newKeys

        
    ### The type of records this family uses
    recordType = Record


    ### The record cache - can be overridden w/any cache type in subclasses

    from TW.Caching import WeakCache as recordCache














    def _getType(self, typeName):
        # XXX check interfaces?  check if contained in self?
        return self.getService(typeName)
        

    def _getKeysFor(self, mapping):

        keys = []
        g = mapping.get

        for key in self.allKeys:

            values = []

            for k in key:
                v = g(k)
                if v is None:
                    break
                values.append(key)
            else:
                keys.append( (key,tuple(values)) )

        return keys


















    ### Cache Reading and Writing

    def _getFromCache(self, keyMapping):
    
        keys = self._getKeysFor(keyMapping)
        
        if not keys:
            raise KeyError, ("No unique keys supplied",keyMapping)
            
        get = self.recordCache.get
        
        for key in keys:
            record = get(key)
            if records is not None:
                return record
        else:
            km = {}
            for fieldNames,fieldValues in keys:
                for f in fieldNames:
                    km[f]=keyMapping[f]

            record = self.recordType(self.getSEFparent(), self, **km)
            self._cacheRecord(record)
            return record


    def getItem(self, **keyMapping):
    
        """Look up record using keyword arguments
        
            Note that this may return 'nonexistent' records; i.e. you must
            call the returned record's exists() or hasType() method if you
            want to be sure it exists."""
            
        return self._getFromCache(keymapping)






class SimpleTypeManager(RecordType, TypeManager):

    """A TypeManager+RecordType combo for type families w/only one type"""

    __implements__ = ITypeImpl, ITypeManager

    rootTypeName = None

    typeName = None
    
    def _getType(self, typeName):
        assert typeName is None, (
            "SimpleTypeManager %s has no subtypes" % self.__class__.__name__
        )
        return self




setupModule()
