"""Transaction Management, w/support for read-only vs. read-write transactions

    This is just a thin shim over the normal Zope 2/ZPublisher/ZODB transaction
    interfaces, adding support for declaring a transaction pure read-only, or
    read-write.  Read-write transactions must be *explicitly declared*, or they
    will not be allowed to make changes.  Declaring a transaction read-only
    blocks it from being declared read-write.  Transactions of unspecified mode
    can perform reads, but may lose data from their per-transaction caches if
    the transaction is later declared read-write (since highly volatile data
    may have been changed by another transaction, before the transaction was
    declared read-write).

    It's not fully implemented yet, however.  Notably, the flushing of cached
    volatile data is not yet supported.
"""

from TW.API import *
from TW.Database.Interfaces import *
from TW.Caching import PermanentCache

__all__ = [
    'TransactionManager', 'TransactionalCache', 'TransactionNotifier',
    'TXN_READONLY', 'TXN_READWRITE', 'TXN_UNDECLARED',
]


TXN_READONLY     = 'ReadOnly'
TXN_READWRITE    = 'ReadWrite'
TXN_UNDECLARED   = None












class TransactionNotifier:
    
    """Transaction boundary helper object"""

    __metaclass__ = OnceClass

    def __init__(self, client, d, name):

        self.name = name

        try:    t = get_transaction()
        except: pass
        else:   t.register(self); self.client = client

    def txnDone(self,*args):
        self.client.txnDone()
        delattr(self.client,self.name)

    # tpc_begin, tpc_vote, commit, commit_sub: ignore
    # tpc_finish, tpc_abort, abort, abort_sub: txn done

    tpc_finish = tpc_abort = abort = abort_sub = txnDone

    def tpc_begin(self,transaction,subtransaction=None): pass
    def commit(self,object,transaction): pass
    def tpc_vote(self,transaction):      pass
    def commit_sub(self,transaction):    pass














class TransactionalCache(PermanentCache):

    """Clears cache on (sub-)transaction boundaries"""

    notifier = TransactionNotifier

    def __setitem__(self, key, value):
        self.notifier   # ensure that we're registered, then proceed
        super(TransactionalCache,self).__setitem__(key, value)

    def txnDone(self):
        self.clear()





























class TransactionManager(SEF.Service):

    """Track transaction count and mode"""

    
    def writeCounter(self, d, n):
        return [0]

    writeCounter = Once(writeCounter)


    def transactionMode(self, d, n):
        self.notifier   # force registration
        return TXN_UNDECLARED

    transactionMode = Once(transactionMode)


    notifier = TransactionNotifier

    def txnDone(self):
        del self.transactionMode



















    def allowWrites(self):

        mode = self.transactionMode

        if mode is TXN_READWRITE:
            return

        if mode is TXN_READONLY:
            raise TransactionReadOnly
            
        self.transactionMode  = TXN_READWRITE
        self.writeCounter[0] += 1 


    def disableWrites(self):
        if self.transactionMode is not TXN_UNDECLARED:
            return
        self.transactionMode = TXN_READONLY


    def attemptWrite(self):
        if self.transactionMode is not TXN_READWRITE:
            raise ReadWriteNotDeclared


    def register(self, object):
        if self.transactionMode is not TXN_READWRITE:
            raise ReadWriteNotDeclared

        try:    t = get_transaction()
        except: pass
        else:   t.register(object)









    def begin(self):
        try:    t = get_transaction()
        except: pass
        else:   t.begin()

    def commit(self, subtxn=0):
        try:    t = get_transaction()
        except: pass
        else:   t.commit(subtxn)

    def abort(self, subtxn=0):
        try:    t = get_transaction()
        except: pass
        else:   t.abort(subtxn)

    def recordMetaData(self, object, request):
        pass    # XXX ???



setupModule()









    

